#!/usr/bin/env python3

"""
Functions for computing the maximum r at which the flux surfaces
become singular.
"""

import logging
import warnings
import numpy as np
#from .util import Struct, fourier_minimum

#logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

def calculate_r_singularity(self, high_order=False):
    """
    """

    # Shorthand:
    s = self
    
    X1c = s.X1c
    Y1s = s.Y1s
    Y1c = s.Y1c

    X20 = s.X20
    X2s = s.X2s
    X2c = s.X2c

    Y20 = s.Y20
    Y2s = s.Y2s
    Y2c = s.Y2c

    Z20 = s.Z20
    Z2s = s.Z2s
    Z2c = s.Z2c

    iota_N0 = s.iotaN
    iota = s.iota
    lp = np.abs(s.G0) / s.B0

    curvature = s.curvature
    torsion = s.torsion

    nphi = s.nphi
    sign_G = s.sG
    sign_psi = s.spsi
    B0 = s.B0
    G0 = s.G0
    I2 = s.I2
    G2 = s.G2
    p2 = s.p2

    B20 = s.B20
    B2s = s.B2s
    B2c = s.B2c

    d_X1c_d_varphi = s.d_X1c_d_varphi
    d_Y1s_d_varphi = s.d_Y1s_d_varphi
    d_Y1c_d_varphi = s.d_Y1c_d_varphi

    d_X20_d_varphi = s.d_X20_d_varphi
    d_X2s_d_varphi = s.d_X2s_d_varphi
    d_X2c_d_varphi = s.d_X2c_d_varphi

    d_Y20_d_varphi = s.d_Y20_d_varphi
    d_Y2s_d_varphi = s.d_Y2s_d_varphi
    d_Y2c_d_varphi = s.d_Y2c_d_varphi

    d_Z20_d_varphi = s.d_Z20_d_varphi
    d_Z2s_d_varphi = s.d_Z2s_d_varphi
    d_Z2c_d_varphi = s.d_Z2c_d_varphi

    d2_X1c_d_varphi2 = s.d2_X1c_d_varphi2
    d2_Y1s_d_varphi2 = s.d2_Y1s_d_varphi2
    d2_Y1c_d_varphi2 = s.d2_Y1c_d_varphi2

    d_curvature_d_varphi = s.d_curvature_d_varphi
    d_torsion_d_varphi = s.d_torsion_d_varphi

    r_singularity_basic_vs_varphi = np.zeros(nphi)
    r_singularity_vs_varphi = np.zeros(nphi)
    r_singularity_residual_sqnorm = np.zeros(nphi)
    r_singularity_theta_vs_varphi = np.zeros(nphi)
    
    # Write sqrt(g) = r * [g0 + r*g1c*cos(theta) + (r^2)*(g20 + g2s*sin(2*theta) + g2c*cos(2*theta) + ...]
    # The coefficients are evaluated in "20200322-02 Max r for Garren Boozer.nb", in the section "Order r^2 construction, quasisymmetry"

    g0 = lp * X1c * Y1s

    #g1s = -2*X20*Y1c + 2*X2c*Y1c + 2*X2s*Y1s + 2*X1c*Y20 - 2*X1c*Y2c
    # g1s vanishes for quasisymmetry.

    g1c = lp*(-2*X2s*Y1c + 2*X20*Y1s + 2*X2c*Y1s + 2*X1c*Y2s - X1c*X1c*Y1s*curvature)

    g20 = -4*lp*X2s*Y2c + 4*lp*X2c*Y2s + lp*X1c*X2s*Y1c*curvature - \
        2*lp*X1c*X20*Y1s*curvature - lp*X1c*X2c*Y1s*curvature - \
        lp*X1c*X1c*Y2s*curvature + 2*lp*Y1c*Y1s*Z2c*torsion - \
        lp*X1c*X1c*Z2s*torsion - lp*Y1c*Y1c*Z2s*torsion + lp*Y1s*Y1s*Z2s*torsion - \
        Y1s*Z20*d_X1c_d_varphi - Y1s*Z2c*d_X1c_d_varphi + \
        Y1c*Z2s*d_X1c_d_varphi - X1c*Z2s*d_Y1c_d_varphi - \
        X1c*Z20*d_Y1s_d_varphi + X1c*Z2c*d_Y1s_d_varphi + \
        X1c*Y1s*d_Z20_d_varphi
    
    g2c = -4*lp*X2s*Y20 + 4*lp*X20*Y2s + \
        lp*X1c*X2s*Y1c*curvature - lp*X1c*X20*Y1s*curvature - \
        2*lp*X1c*X2c*Y1s*curvature - lp*X1c*X1c*Y2s*curvature + \
        2*lp*Y1c*Y1s*Z20*torsion - lp*X1c*X1c*Z2s*torsion - \
        lp*Y1c*Y1c*Z2s*torsion - lp*Y1s*Y1s*Z2s*torsion - \
        Y1s*Z20*d_X1c_d_varphi - Y1s*Z2c*d_X1c_d_varphi + \
        Y1c*Z2s*d_X1c_d_varphi - X1c*Z2s*d_Y1c_d_varphi + \
        X1c*Z20*d_Y1s_d_varphi - X1c*Z2c*d_Y1s_d_varphi + \
        X1c*Y1s*d_Z2c_d_varphi
    
    g2s = 4*lp*X2c*Y20 - 4*lp*X20*Y2c + \
        lp*X1c*X20*Y1c*curvature - lp*X1c*X2c*Y1c*curvature - \
        2*lp*X1c*X2s*Y1s*curvature - lp*X1c*X1c*Y20*curvature + \
        lp*X1c*X1c*Y2c*curvature - lp*X1c*X1c*Z20*torsion - \
        lp*Y1c*Y1c*Z20*torsion + lp*Y1s*Y1s*Z20*torsion + \
        lp*X1c*X1c*Z2c*torsion + lp*Y1c*Y1c*Z2c*torsion + \
        lp*Y1s*Y1s*Z2c*torsion + Y1c*Z20*d_X1c_d_varphi - \
        Y1c*Z2c*d_X1c_d_varphi - Y1s*Z2s*d_X1c_d_varphi - \
        X1c*Z20*d_Y1c_d_varphi + X1c*Z2c*d_Y1c_d_varphi - \
        X1c*Z2s*d_Y1s_d_varphi + X1c*Y1s*d_Z2s_d_varphi
    
    if high_order:
        g3s1 = lp*(2*X20*X20*Y1c*curvature + X2c*X2c*Y1c*curvature + X2s*X2s*Y1c*curvature - X1c*X2s*Y2s*curvature + \
                   2*Y1c*Z20*Z20*curvature - 3*Y1c*Z20*Z2c*curvature + Y1c*Z2c*Z2c*curvature - 3*Y1s*Z20*Z2s*curvature + \
                   Y1c*Z2s*Z2s*curvature - 2*Y1c*Y20*Z20*torsion - Y1c*Y2c*Z20*torsion - Y1s*Y2s*Z20*torsion + \
                   4*Y1c*Y20*Z2c*torsion - Y1c*Y2c*Z2c*torsion + 5*Y1s*Y2s*Z2c*torsion - \
                   X1c*X2s*Z2s*torsion + 4*Y1s*Y20*Z2s*torsion - 5*Y1s*Y2c*Z2s*torsion - \
                   Y1c*Y2s*Z2s*torsion - X1c*X2c*(Y20*curvature + Y2c*curvature + (Z20 + Z2c)*torsion) - \
                   X20*(3*X2c*Y1c*curvature + 3*X2s*Y1s*curvature + \
                        2*X1c*(Y20*curvature - 2*Y2c*curvature + (Z20 - 2*Z2c)*torsion))) - 2*Y20*Z2c*d_X1c_d_varphi + \
                        2*Y1c*Z20*d_X20_d_varphi - 2*Y1c*Z2c*d_X20_d_varphi - \
                        2*Y1s*Z2s*d_X20_d_varphi - Y1c*Z20*d_X2c_d_varphi + Y1c*Z2c*d_X2c_d_varphi + \
                        Y1s*Z2s*d_X2c_d_varphi - Y1s*Z20*d_X2s_d_varphi - Y1s*Z2c*d_X2s_d_varphi + \
                        Y1c*Z2s*d_X2s_d_varphi - 2*X2c*Z20*d_Y1c_d_varphi + \
                        2*X20*Z2c*d_Y1c_d_varphi - 2*X2s*Z20*d_Y1s_d_varphi + \
                        4*X2s*Z2c*d_Y1s_d_varphi + 2*X20*Z2s*d_Y1s_d_varphi - \
                        4*X2c*Z2s*d_Y1s_d_varphi - 2*X1c*Z20*d_Y20_d_varphi + \
                        2*X1c*Z2c*d_Y20_d_varphi + X1c*Z20*d_Y2c_d_varphi - X1c*Z2c*d_Y2c_d_varphi - \
                        X1c*Z2s*d_Y2s_d_varphi - 2*X20*Y1c*d_Z20_d_varphi + \
                        2*X2c*Y1c*d_Z20_d_varphi + 2*X2s*Y1s*d_Z20_d_varphi + \
                        2*X1c*Y20*d_Z20_d_varphi + X20*Y1c*d_Z2c_d_varphi - X2c*Y1c*d_Z2c_d_varphi - \
                        X2s*Y1s*d_Z2c_d_varphi - X1c*Y20*d_Z2c_d_varphi + \
                        Y2c*(2*Z20*d_X1c_d_varphi + X1c*(-2*d_Z20_d_varphi + d_Z2c_d_varphi)) - \
                        X2s*Y1c*d_Z2s_d_varphi + X20*Y1s*d_Z2s_d_varphi + X2c*Y1s*d_Z2s_d_varphi + \
                        X1c*Y2s*d_Z2s_d_varphi
        
        g3s3 = lp*(-(X2c*X2c*Y1c*curvature) + X2s*X2s*Y1c*curvature - X1c*X2s*Y2s*curvature + Y1c*Z20*Z2c*curvature - \
                   Y1c*Z2c*Z2c*curvature - Y1s*Z20*Z2s*curvature - 2*Y1s*Z2c*Z2s*curvature + Y1c*Z2s*Z2s*curvature - \
                   3*Y1c*Y2c*Z20*torsion + 3*Y1s*Y2s*Z20*torsion + 2*Y1c*Y20*Z2c*torsion + \
                   Y1c*Y2c*Z2c*torsion + Y1s*Y2s*Z2c*torsion - X1c*X2s*Z2s*torsion - \
                   2*Y1s*Y20*Z2s*torsion + Y1s*Y2c*Z2s*torsion - Y1c*Y2s*Z2s*torsion + \
                   X20*(X2c*Y1c*curvature - X2s*Y1s*curvature + 2*X1c*(Y2c*curvature + Z2c*torsion)) + \
                   X2c*(-2*X2s*Y1s*curvature + X1c*(-3*Y20*curvature + Y2c*curvature + (-3*Z20 + Z2c)*torsion))) - \
                   2*Y20*Z2c*d_X1c_d_varphi + Y1c*Z20*d_X2c_d_varphi - Y1c*Z2c*d_X2c_d_varphi - \
                   Y1s*Z2s*d_X2c_d_varphi - Y1s*Z20*d_X2s_d_varphi - Y1s*Z2c*d_X2s_d_varphi + \
                   Y1c*Z2s*d_X2s_d_varphi - 2*X2c*Z20*d_Y1c_d_varphi + \
                   2*X20*Z2c*d_Y1c_d_varphi + 2*X2s*Z20*d_Y1s_d_varphi - \
                   2*X20*Z2s*d_Y1s_d_varphi - X1c*Z20*d_Y2c_d_varphi + X1c*Z2c*d_Y2c_d_varphi - \
                   X1c*Z2s*d_Y2s_d_varphi - X20*Y1c*d_Z2c_d_varphi + X2c*Y1c*d_Z2c_d_varphi + \
                   X2s*Y1s*d_Z2c_d_varphi + X1c*Y20*d_Z2c_d_varphi + \
                   Y2c*(2*Z20*d_X1c_d_varphi - X1c*d_Z2c_d_varphi) - X2s*Y1c*d_Z2s_d_varphi + \
                   X20*Y1s*d_Z2s_d_varphi + X2c*Y1s*d_Z2s_d_varphi + X1c*Y2s*d_Z2s_d_varphi
        
        g3c1 = -(lp*(2*X20*X20*Y1s*curvature + X2c*X2c*Y1s*curvature + X2s*X2s*Y1s*curvature - X1c*X2s*Y20*curvature - \
                     5*X1c*X2s*Y2c*curvature + 2*Y1s*Z20*Z20*curvature + 3*Y1s*Z20*Z2c*curvature + Y1s*Z2c*Z2c*curvature - \
                     3*Y1c*Z20*Z2s*curvature + Y1s*Z2s*Z2s*curvature - X1c*X2s*Z20*torsion - \
                     2*Y1s*Y20*Z20*torsion + Y1s*Y2c*Z20*torsion - Y1c*Y2s*Z20*torsion - \
                     5*X1c*X2s*Z2c*torsion - 4*Y1s*Y20*Z2c*torsion - Y1s*Y2c*Z2c*torsion - \
                     5*Y1c*Y2s*Z2c*torsion + 4*Y1c*Y20*Z2s*torsion + 5*Y1c*Y2c*Z2s*torsion - \
                     Y1s*Y2s*Z2s*torsion + 5*X1c*X2c*(Y2s*curvature + Z2s*torsion) + \
                     X20*(-3*X2s*Y1c*curvature + 3*X2c*Y1s*curvature + 4*X1c*(Y2s*curvature + Z2s*torsion)))) + \
                     2*Y20*Z2s*d_X1c_d_varphi + 4*Y2c*Z2s*d_X1c_d_varphi - \
                     2*Y1s*Z20*d_X20_d_varphi - 2*Y1s*Z2c*d_X20_d_varphi + \
                     2*Y1c*Z2s*d_X20_d_varphi - Y1s*Z20*d_X2c_d_varphi - Y1s*Z2c*d_X2c_d_varphi + \
                     Y1c*Z2s*d_X2c_d_varphi + Y1c*Z20*d_X2s_d_varphi - Y1c*Z2c*d_X2s_d_varphi - \
                     Y1s*Z2s*d_X2s_d_varphi + 2*X2s*Z20*d_Y1c_d_varphi + \
                     4*X2s*Z2c*d_Y1c_d_varphi - 2*X20*Z2s*d_Y1c_d_varphi - \
                     4*X2c*Z2s*d_Y1c_d_varphi - 2*X2c*Z20*d_Y1s_d_varphi + \
                     2*X20*Z2c*d_Y1s_d_varphi - 2*X1c*Z2s*d_Y20_d_varphi - \
                     X1c*Z2s*d_Y2c_d_varphi - X1c*Z20*d_Y2s_d_varphi + X1c*Z2c*d_Y2s_d_varphi - \
                     2*X2s*Y1c*d_Z20_d_varphi + 2*X20*Y1s*d_Z20_d_varphi + \
                     2*X2c*Y1s*d_Z20_d_varphi - X2s*Y1c*d_Z2c_d_varphi + X20*Y1s*d_Z2c_d_varphi + \
                     X2c*Y1s*d_Z2c_d_varphi + Y2s*\
                     (-2*Z20*d_X1c_d_varphi - 4*Z2c*d_X1c_d_varphi + \
                      X1c*(2*d_Z20_d_varphi + d_Z2c_d_varphi)) - X20*Y1c*d_Z2s_d_varphi + \
                      X2c*Y1c*d_Z2s_d_varphi + X2s*Y1s*d_Z2s_d_varphi + X1c*Y20*d_Z2s_d_varphi - \
                      X1c*Y2c*d_Z2s_d_varphi
        
        g3c3 = -(lp*(X2c*X2c*Y1s*curvature - X2s*X2s*Y1s*curvature - 3*X1c*X2s*Y20*curvature + X1c*X2s*Y2c*curvature + \
                     Y1s*Z20*Z2c*curvature + Y1s*Z2c*Z2c*curvature + Y1c*Z20*Z2s*curvature - 2*Y1c*Z2c*Z2s*curvature - \
                     Y1s*Z2s*Z2s*curvature - 3*X1c*X2s*Z20*torsion - 3*Y1s*Y2c*Z20*torsion - \
                     3*Y1c*Y2s*Z20*torsion + X1c*X2s*Z2c*torsion + 2*Y1s*Y20*Z2c*torsion - \
                     Y1s*Y2c*Z2c*torsion + Y1c*Y2s*Z2c*torsion + 2*Y1c*Y20*Z2s*torsion + \
                     Y1c*Y2c*Z2s*torsion + Y1s*Y2s*Z2s*torsion + \
                     X2c*(-2*X2s*Y1c*curvature + X1c*(Y2s*curvature + Z2s*torsion)) + \
                     X20*(X2s*Y1c*curvature + X2c*Y1s*curvature + 2*X1c*(Y2s*curvature + Z2s*torsion)))) + \
                     2*Y20*Z2s*d_X1c_d_varphi - Y1s*Z20*d_X2c_d_varphi - Y1s*Z2c*d_X2c_d_varphi + \
                     Y1c*Z2s*d_X2c_d_varphi - Y1c*Z20*d_X2s_d_varphi + Y1c*Z2c*d_X2s_d_varphi + \
                     Y1s*Z2s*d_X2s_d_varphi + 2*X2s*Z20*d_Y1c_d_varphi - \
                     2*X20*Z2s*d_Y1c_d_varphi + 2*X2c*Z20*d_Y1s_d_varphi - \
                     2*X20*Z2c*d_Y1s_d_varphi - X1c*Z2s*d_Y2c_d_varphi + X1c*Z20*d_Y2s_d_varphi - \
                     X1c*Z2c*d_Y2s_d_varphi - X2s*Y1c*d_Z2c_d_varphi + X20*Y1s*d_Z2c_d_varphi + \
                     X2c*Y1s*d_Z2c_d_varphi + Y2s*(-2*Z20*d_X1c_d_varphi + X1c*d_Z2c_d_varphi) + \
                     X20*Y1c*d_Z2s_d_varphi - X2c*Y1c*d_Z2s_d_varphi - X2s*Y1s*d_Z2s_d_varphi - \
                     X1c*Y20*d_Z2s_d_varphi + X1c*Y2c*d_Z2s_d_varphi
        
        g40 = -2*(-3*lp*(-((Y2s*Z2c - Y2c*Z2s)*(Z20*curvature - Y20*torsion)) + \
                         X20*(X2s*(Y2c*curvature + Z2c*torsion) - X2c*(Y2s*curvature + Z2s*torsion))) - \
                  2*Y2c*Z2s*d_X20_d_varphi - Y20*Z2s*d_X2c_d_varphi - \
                  Y2c*Z20*d_X2s_d_varphi + Y20*Z2c*d_X2s_d_varphi - \
                  2*X2s*Z2c*d_Y20_d_varphi + 2*X2c*Z2s*d_Y20_d_varphi - \
                  X2s*Z20*d_Y2c_d_varphi + X20*Z2s*d_Y2c_d_varphi + X2c*Z20*d_Y2s_d_varphi - \
                  X20*Z2c*d_Y2s_d_varphi + 2*X2s*Y2c*d_Z20_d_varphi + \
                  X2s*Y20*d_Z2c_d_varphi + Y2s*\
                  (2*Z2c*d_X20_d_varphi + Z20*d_X2c_d_varphi - 2*X2c*d_Z20_d_varphi - \
                   X20*d_Z2c_d_varphi) - X2c*Y20*d_Z2s_d_varphi + X20*Y2c*d_Z2s_d_varphi)
        
        g4s2 = 4*(lp*(Y2c*Z20*Z20*curvature - Y20*Z20*Z2c*curvature - Y2s*Z2c*Z2s*curvature + Y2c*Z2s*Z2s*curvature - \
                      Y20*Y2c*Z20*torsion + Y20*Y20*Z2c*torsion + Y2s*Y2s*Z2c*torsion - Y2c*Y2s*Z2s*torsion - \
                      X20*X2c*(Y20*curvature + Z20*torsion) + X20*X20*(Y2c*curvature + Z2c*torsion) + \
                      X2s*X2s*(Y2c*curvature + Z2c*torsion) - X2c*X2s*(Y2s*curvature + Z2s*torsion)) - \
                  Y20*Z2c*d_X20_d_varphi - Y2s*Z2c*d_X2s_d_varphi - X2c*Z20*d_Y20_d_varphi + \
                  X20*Z2c*d_Y20_d_varphi + X2s*Z2c*d_Y2s_d_varphi - X2c*Z2s*d_Y2s_d_varphi + \
                  X2c*Y20*d_Z20_d_varphi + X2c*Y2s*d_Z2s_d_varphi + \
                  Y2c*(Z20*d_X20_d_varphi + Z2s*d_X2s_d_varphi - X20*d_Z20_d_varphi - \
                       X2s*d_Z2s_d_varphi))
        
        g4s4 = 2*(lp*(Y2c*Z20*Z2c*curvature - Y20*Z2c*Z2c*curvature - Y2s*Z20*Z2s*curvature + Y20*Z2s*Z2s*curvature - \
                      Y2c*Y2c*Z20*torsion + Y2s*Y2s*Z20*torsion + Y20*Y2c*Z2c*torsion - Y20*Y2s*Z2s*torsion - \
                      X2c*X2c*(Y20*curvature + Z20*torsion) + X2s*X2s*(Y20*curvature + Z20*torsion) + \
                      X20*X2c*(Y2c*curvature + Z2c*torsion) - X20*X2s*(Y2s*curvature + Z2s*torsion)) - \
                  Y20*Z2c*d_X2c_d_varphi - Y2s*Z20*d_X2s_d_varphi + Y20*Z2s*d_X2s_d_varphi - \
                  X2c*Z20*d_Y2c_d_varphi + X20*Z2c*d_Y2c_d_varphi + X2s*Z20*d_Y2s_d_varphi - \
                  X20*Z2s*d_Y2s_d_varphi + X2c*Y20*d_Z2c_d_varphi + \
                  Y2c*(Z20*d_X2c_d_varphi - X20*d_Z2c_d_varphi) - X2s*Y20*d_Z2s_d_varphi + \
                  X20*Y2s*d_Z2s_d_varphi)
        
        g4c2 = -4*(lp*(Y2s*Z20*Z20*curvature + Y2s*Z2c*Z2c*curvature - Y20*Z20*Z2s*curvature - Y2c*Z2c*Z2s*curvature - \
                       Y20*Y2s*Z20*torsion - Y2c*Y2s*Z2c*torsion + Y20*Y20*Z2s*torsion + Y2c*Y2c*Z2s*torsion - \
                       X20*X2s*(Y20*curvature + Z20*torsion) - X2c*X2s*(Y2c*curvature + Z2c*torsion) + \
                       X20*X20*(Y2s*curvature + Z2s*torsion) + X2c*X2c*(Y2s*curvature + Z2s*torsion)) - \
                   Y20*Z2s*d_X20_d_varphi - Y2c*Z2s*d_X2c_d_varphi - X2s*Z20*d_Y20_d_varphi + \
                   X20*Z2s*d_Y20_d_varphi - X2s*Z2c*d_Y2c_d_varphi + X2c*Z2s*d_Y2c_d_varphi + \
                   X2s*Y20*d_Z20_d_varphi + X2s*Y2c*d_Z2c_d_varphi + \
                   Y2s*(Z20*d_X20_d_varphi + Z2c*d_X2c_d_varphi - X20*d_Z20_d_varphi - \
                        X2c*d_Z2c_d_varphi))
        
        g4c4 = -2*(lp*(Y2s*Z20*Z2c*curvature + Y2c*Z20*Z2s*curvature - 2*Y20*Z2c*Z2s*curvature - \
                       2*Y2c*Y2s*Z20*torsion + Y20*Y2s*Z2c*torsion + Y20*Y2c*Z2s*torsion + \
                       X20*X2s*(Y2c*curvature + Z2c*torsion) + \
                       X2c*(-2*X2s*(Y20*curvature + Z20*torsion) + X20*(Y2s*curvature + Z2s*torsion))) - \
                   Y20*Z2s*d_X2c_d_varphi + Y2c*Z20*d_X2s_d_varphi - Y20*Z2c*d_X2s_d_varphi - \
                   X2s*Z20*d_Y2c_d_varphi + X20*Z2s*d_Y2c_d_varphi - X2c*Z20*d_Y2s_d_varphi + \
                   X20*Z2c*d_Y2s_d_varphi + X2s*Y20*d_Z2c_d_varphi + \
                   Y2s*(Z20*d_X2c_d_varphi - X20*d_Z2c_d_varphi) + X2c*Y20*d_Z2s_d_varphi - \
                   X20*Y2c*d_Z2s_d_varphi)


    # We consider the system sqrt(g) = 0 and
    # d (sqrtg) / d theta = 0.
    # We algebraically eliminate r in "20200322-02 Max r for Garren Boozer.nb", in the section
    # "Keeping first 3 orders in the Jacobian".
    # We end up with the form in "20200322-01 Max r for GarrenBoozer.docx":
    # K0 + K2s*sin(2*theta) + K2c*cos(2*theta) + K4s*sin(4*theta) + K4c*cos(4*theta) = 0.

    K0 = 2*g1c*g1c*g20 - 3*g1c*g1c*g2c + 8*g0*g2c*g2c + 8*g0*g2s*g2s

    K2s = 2*g1c*g1c*g2s

    K2c = -2*g1c*g1c*g20 + 2*g1c*g1c*g2c

    K4s = g1c*g1c*g2s - 16*g0*g2c*g2s

    K4c = g1c*g1c*g2c - 8*g0*g2c*g2c + 8*g0*g2s*g2s

    coefficients = np.zeros((nphi,5))
    
    coefficients[:, 4] = 4*(K4c*K4c + K4s*K4s)

    coefficients[:, 3] = 4*(K4s*K2c - K2s*K4c)

    coefficients[:, 2] = K2s*K2s + K2c*K2c - 4*K0*K4c - 4*K4c*K4c - 4*K4s*K4s

    coefficients[:, 1] = 2*K0*K2s + 2*K4c*K2s - 4*K4s*K2c

    coefficients[:, 0] = (K0 + K4c)*(K0 + K4c) - K2c*K2c

    for jphi in range(nphi):
        # Solve for the roots of the quartic polynomial:
        try:
            roots = np.polynomial.polynomial.polyroots(coefficients[jphi, :]) # Do I need to reverse the order of the coefficients?
        except np.linalg.LinAlgError:
            raise RuntimeError('Problem with polyroots. coefficients={} lp={} B0={} g0={} g1c={}'.format(coefficients[jphi, :], lp, s.B0, g0, g1c))

        real_parts = np.real(roots)
        imag_parts = np.imag(roots)

        logger.debug('jphi={} g0={} g1c={} g20={} g2s={} g2c={} K0={} K2s={} K2c={} K4s={} K4c={} coefficients={} real={} imag={}'.format(jphi, g0[jphi], g1c[jphi], g20[jphi], g2s[jphi], g2c[jphi], K0[jphi], K2s[jphi], K2c[jphi], K4s[jphi], K4c[jphi], coefficients[jphi,:], real_parts, imag_parts))

        # This huge number indicates a true solution has not yet been found.
        rc = 1e+100

        for jr in range(4):
            # Loop over the roots of the equation for w.

            # If root is not purely real, skip it.
            if np.abs(imag_parts[jr]) > 1e-7:
                logger.debug("Skipping root with jr={} since imag part is {}".format(jr, imag_parts[jr]))
                continue

            sin2theta = real_parts[jr]

            # Discard any roots that have magnitude larger than 1. (I'm not sure this ever happens, but check to be sure.)
            if np.abs(sin2theta) > 1:
                logger.debug("Skipping root with jr={} since sin2theta={}".format(jr, sin2theta))
                continue

            # Determine varpi by checking which choice gives the smaller residual in the K equation
            abs_cos2theta = np.sqrt(1 - sin2theta * sin2theta)
            residual_if_varpi_plus  = np.abs(K0[jphi] + K2s[jphi] * sin2theta + K2c[jphi] *   abs_cos2theta \
                                             + K4s[jphi] * 2 * sin2theta *   abs_cos2theta  + K4c[jphi] * (1 - 2 * sin2theta * sin2theta))
            residual_if_varpi_minus = np.abs(K0[jphi] + K2s[jphi] * sin2theta + K2c[jphi] * (-abs_cos2theta) \
                                             + K4s[jphi] * 2 * sin2theta * (-abs_cos2theta) + K4c[jphi] * (1 - 2 * sin2theta * sin2theta))

            if residual_if_varpi_plus > residual_if_varpi_minus:
                varpi = -1
            else:
                varpi = 1

            cos2theta = varpi * abs_cos2theta

            # The next few lines give an older method for computing varpi, which has problems in edge cases
            # where w (the root of the quartic polynomial) is very close to +1 or -1, giving varpi
            # not very close to +1 or -1 due to bad loss of precision.
            #
            #varpi_denominator = ((K4s*2*sin2theta + K2c) * sqrt(1 - sin2theta*sin2theta))
            #if (abs(varpi_denominator) < 1e-8) print *,"WARNING!!! varpi_denominator=",varpi_denominator
            #varpi = -(K0 + K2s * sin2theta + K4c*(1 - 2*sin2theta*sin2theta)) / varpi_denominator
            #if (abs(varpi*varpi-1) > 1e-3) print *,"WARNING!!! abs(varpi*varpi-1) =",abs(varpi*varpi-1)
            #varpi = nint(varpi) ! Ensure varpi is exactly either +1 or -1.
            #cos2theta = varpi * sqrt(1 - sin2theta*sin2theta)

            # To get (sin theta, cos theta) from (sin 2 theta, cos 2 theta), we consider two cases to
            # avoid precision loss when cos2theta is added to or subtracted from 1:
            get_cos_from_cos2 = cos2theta > 0
            if get_cos_from_cos2:
                abs_costheta = np.sqrt(0.5*(1 + cos2theta))
            else:
                abs_sintheta = np.sqrt(0.5 * (1 - cos2theta))
                
            logger.debug("  jr={}  sin2theta={}  cos2theta={}".format(jr, sin2theta, cos2theta))
            for varsigma in [-1, 1]:
                if get_cos_from_cos2:
                    costheta = varsigma * abs_costheta
                    sintheta = sin2theta / (2 * costheta)
                else:
                    sintheta = varsigma * abs_sintheta
                    costheta = sin2theta / (2 * sintheta)
                logger.debug("    varsigma={}  costheta={}  sintheta={}".format(varsigma, costheta, sintheta))

                # Sanity test
                if np.abs(costheta*costheta + sintheta*sintheta - 1) > 1e-13:
                    msg = "Error! sintheta={} costheta={} jphi={} jr={} sin2theta={} cos2theta={} abs(costheta*costheta + sintheta*sintheta - 1)={}".format(sintheta, costheta, jphi, jr, sin2theta, cos2theta, np.abs(costheta*costheta + sintheta*sintheta - 1))
                    logger.error(msg)
                    raise RuntimeError(msg)

                """
                # Try to get r using the simpler method, the equation that is linear in r.
                denominator = 2 * (g2s[jphi] * cos2theta - g2c[jphi] * sin2theta)
                if np.abs(denominator) > 1e-8:
                    # This method cannot be used if we would need to divide by 0
                    rr = g1c[jphi] * sintheta / denominator
                    residual = g0[jphi] + rr * g1c[jphi] * costheta + rr * rr * (g20[jphi] + g2s[jphi] * sin2theta + g2c[jphi] * cos2theta) # Residual in the equation sqrt(g)=0.
                    logger.debug("    Linear method: rr={}  residual={}".format(rr, residual))
                    if (rr>0) and np.abs(residual) < 1e-5:
                        if rr < rc:
                            # If this is a new minimum
                            rc = rr
                            sintheta_at_rc = sintheta
                            costheta_at_rc = costheta
                            logger.debug("      New minimum: rc={}".format(rc))
                else:
                    # Use the more complicated method to determine rr by solving a quadratic equation.
                    quadratic_A = g20[jphi] + g2s[jphi] * sin2theta + g2c[jphi] * cos2theta
                    quadratic_B = costheta * g1c[jphi]
                    quadratic_C = g0[jphi]
                    radical = np.sqrt(quadratic_B * quadratic_B - 4 * quadratic_A * quadratic_C)
                    for sign_quadratic in [-1, 1]:
                        rr = (-quadratic_B + sign_quadratic * radical) / (2 * quadratic_A) # This is the quadratic formula.
                        residual = -g1c[jphi] * sintheta + 2 * rr * (g2s[jphi] * cos2theta - g2c[jphi] * sin2theta) # Residual in the equation d sqrt(g) / d theta = 0.
                        logger.debug("    Quadratic method: A={} B={} C={} radicand={}, radical={}  rr={}  residual={}".format(quadratic_A, quadratic_B, quadratic_C, quadratic_B * quadratic_B - 4 * quadratic_A * quadratic_C, radical, rr, residual))
                        if (rr>0) and np.abs(residual) < 1e-5:
                            if rr < rc:
                                # If this is a new minimum
                                rc = rr
                                sintheta_at_rc = sintheta
                                costheta_at_rc = costheta
                                logger.debug("      New minimum: rc={}".format(rc))
                """

                # Try to get r using the simpler method, the equation that is linear in r.
                linear_solutions = []
                denominator = 2 * (g2s[jphi] * cos2theta - g2c[jphi] * sin2theta)
                if np.abs(denominator) > 1e-8:
                    # This method cannot be used if we would need to divide by 0
                    rr = g1c[jphi] * sintheta / denominator
                    residual = g0[jphi] + rr * g1c[jphi] * costheta + rr * rr * (g20[jphi] + g2s[jphi] * sin2theta + g2c[jphi] * cos2theta) # Residual in the equation sqrt(g)=0.
                    logger.debug("    Linear method: rr={}  residual={}".format(rr, residual))
                    if (rr>0) and np.abs(residual) < 1e-5:
                        linear_solutions = [rr]
                        
                # Use the more complicated method to determine rr by solving a quadratic equation.
                quadratic_solutions = []
                quadratic_A = g20[jphi] + g2s[jphi] * sin2theta + g2c[jphi] * cos2theta
                quadratic_B = costheta * g1c[jphi]
                quadratic_C = g0[jphi]
                radicand = quadratic_B * quadratic_B - 4 * quadratic_A * quadratic_C
                if np.abs(quadratic_A) < 1e-13:
                    rr = -quadratic_C / quadratic_B
                    residual = -g1c[jphi] * sintheta + 2 * rr * (g2s[jphi] * cos2theta - g2c[jphi] * sin2theta) # Residual in the equation d sqrt(g) / d theta = 0.
                    logger.debug("    Quadratic method but A is small: A={} rr={}  residual={}".format(quadratic_A, rr, residual))
                    if rr > 0 and np.abs(residual) < 1e-5:
                        quadratic_solutions.append(rr)
                else:
                    # quadratic_A is nonzero, so we can apply the quadratic formula.
                    # I've seen a case where radicand is <0 due I think to under-resolution in phi.
                    if radicand >= 0:
                        radical = np.sqrt(radicand)
                        for sign_quadratic in [-1, 1]:
                            rr = (-quadratic_B + sign_quadratic * radical) / (2 * quadratic_A) # This is the quadratic formula.
                            residual = -g1c[jphi] * sintheta + 2 * rr * (g2s[jphi] * cos2theta - g2c[jphi] * sin2theta) # Residual in the equation d sqrt(g) / d theta = 0.
                            logger.debug("    Quadratic method: A={} B={} C={} radicand={}, radical={}  rr={}  residual={}".format(quadratic_A, quadratic_B, quadratic_C, quadratic_B * quadratic_B - 4 * quadratic_A * quadratic_C, radical, rr, residual))
                            if (rr>0) and np.abs(residual) < 1e-5:
                                quadratic_solutions.append(rr)

                logger.debug("    # linear solutions={}  # quadratic solutions={}".format(len(linear_solutions), len(quadratic_solutions)))
                if len(quadratic_solutions) > 1:
                    # Pick the smaller one
                    quadratic_solutions = [np.min(quadratic_solutions)]
                    
                # If both methods find a solution, check that they agree:
                if len(linear_solutions) > 0 and len(quadratic_solutions) > 0:
                    diff = np.abs(linear_solutions[0] - quadratic_solutions[0])
                    logger.debug("  linear solution={}  quadratic solution={}  diff={}".format(linear_solutions[0], quadratic_solutions[0], diff))
                    # if diff > 1e-5:
                    #     warnings.warn("  Difference between linear solution {} and quadratic solution {} is {}".format(linear_solutions[0], quadratic_solutions[0], diff))
                        
                    #assert np.abs(linear_solutions[0] - quadratic_solutions[0]) < 1e-5, "Difference between linear solution {} and quadratic solution {} is {}".format(linear_solutions[0], quadratic_solutions[0], linear_solutions[0] - quadratic_solutions[0])
                    
                # Prefer the quadratic solution
                rr = -1
                if len(quadratic_solutions) > 0:
                    rr = quadratic_solutions[0]
                elif len(linear_solutions) > 0:
                    rr = linear_solutions[0]

                if rr > 0 and rr < rc:
                    # This is a new minimum
                    rc = rr
                    sintheta_at_rc = sintheta
                    costheta_at_rc = costheta
                    logger.debug("      New minimum: rc={}".format(rc))
                    
        r_singularity_basic_vs_varphi[jphi] = rc
        #r_singularity_Newton_solve()
        r_singularity_vs_varphi[jphi] = rc
        r_singularity_residual_sqnorm[jphi] = 0 # Newton_residual_sqnorm
        r_singularity_theta_vs_varphi[jphi] = 0 # theta FIX ME!!

    self.r_singularity_vs_varphi = r_singularity_vs_varphi
    self.inv_r_singularity_vs_varphi = 1 / r_singularity_vs_varphi
    self.r_singularity_basic_vs_varphi = r_singularity_basic_vs_varphi
    self.r_singularity = np.min(r_singularity_vs_varphi)    
    self.r_singularity_theta_vs_varphi = r_singularity_theta_vs_varphi
    self.r_singularity_residual_sqnorm = r_singularity_residual_sqnorm
    
"""
  call cpu_time(end_time)
  if (verbose) print "(a,es11.4,a,es10.3,a)"," r_singularity:",r_singularity,"  Time to compute:",end_time - r_singularity_start_time," sec."


contains

  subroutine r_singularity_Newton_solve()
    ! Apply Newton's method to iteratively refine the solution for (r,theta) where the surfaces become singular.

    use quasisymmetry_variables, only: r_singularity_Newton_iterations, r_singularity_line_search, r_singularity_Newton_tolerance, verbose
    implicit none

    real(dp) :: state0(2), step_direction(2), step_scale, last_Newton_residual_sqnorm
    integer :: j_Newton, j_line_search
    logical :: verbose_Newton
    real(dp) :: fd_Jacobian(2,2), state_plus(2), state_minus(2), residual_plus(2), residual_minus(2), delta = 1.0d-8

    verbose_Newton = verbose
    theta = atan2(sintheta_at_rc, costheta_at_rc)
    state(1) = rc
    state(2) = theta

!!$    if (j==80) then
!!$       state0 = state
!!$
!!$       state(1) = state0(1) + delta
!!$       call r_singularity_residual()
!!$       residual_plus = Newton_residual
!!$       state(1) = state0(1) - delta
!!$       call r_singularity_residual()
!!$       residual_minus = Newton_residual
!!$       fd_Jacobian(:,1) = (residual_plus - residual_minus) / (2*delta)
!!$       state = state0
!!$
!!$       state(2) = state0(2) + delta
!!$       call r_singularity_residual()
!!$       residual_plus = Newton_residual
!!$       state(2) = state0(2) - delta
!!$       call r_singularity_residual()
!!$       residual_minus = Newton_residual
!!$       fd_Jacobian(:,2) = (residual_plus - residual_minus) / (2*delta)
!!$       state = state0
!!$
!!$       print *," ZZZ FD Jacobian:"
!!$       print "(2(es24.15))", fd_Jacobian(1,:)
!!$       print "(2(es24.15))", fd_Jacobian(2,:)
!!$    end if

    call r_singularity_residual()

    if (verbose_Newton) print "(a,i4,3(a,es24.15))","r_singularity: jphi=",j," r=",state(1)," th=",state(2)," Residual L2 norm:",Newton_residual_sqnorm
    Newton: do j_Newton = 1, r_singularity_Newton_iterations
       if (verbose_Newton) print "(a,i3)","  Newton iteration ",j_Newton
       last_Newton_residual_sqnorm = Newton_residual_sqnorm
       if (last_Newton_residual_sqnorm < r_singularity_Newton_tolerance) exit Newton
       state0 = state
       call r_singularity_Jacobian()
       step_direction = - matmul(inv_Jacobian, Newton_residual)

       ! Don't let the step get too big:
       step_direction = step_direction * min(1.0, 0.1 / abs(step_direction(2))) ! Max absolute change to theta is 0.1
       step_direction = step_direction * min(1.0, 0.1 * abs(state(1)) / abs(step_direction(1))) ! Max relative change to r is 10%

       step_scale = 1
       line_search: do j_line_search = 1, r_singularity_line_search
          state = state0 + step_scale * step_direction
          call r_singularity_residual()
          if (verbose_Newton) print "(a,i3,3(a,es24.15))","    Line search step",j_line_search,"  r=",state(1)," th=",state(2)," Residual L2 norm:",Newton_residual_sqnorm
          if (Newton_residual_sqnorm < last_Newton_residual_sqnorm) exit line_search

          step_scale = step_scale / 2       
       end do line_search

       if (Newton_residual_sqnorm > last_Newton_residual_sqnorm) then
          if (verbose_Newton) print *,"Line search failed to reduce residual."
          exit Newton
       end if
    end do Newton

    rc = state(1)
    theta = state(2)

  end subroutine r_singularity_Newton_solve

  subroutine r_singularity_residual

    implicit none

    real(dp) :: theta0, r0

    r0 = state(1)
    theta0 = state(2)
    sintheta = sin(theta0)
    costheta = cos(theta0)
    sin2theta = sin(2*theta0)
    cos2theta = cos(2*theta0)
    ! If ghat = sqrt{g}/r,
    ! residual = [ghat; d ghat / d theta]
    Newton_residual(1) = g0 + r0 * g1c * costheta + r0 * r0 * (g20 + g2s * sin2theta + g2c * cos2theta)
    Newton_residual(2) = r0 * (-g1c * sintheta) + 2 * r0 * r0 * (g2s * cos2theta - g2c * sin2theta)

    if (r_singularity_high_order) then
       sin3theta = sin(3*theta0)
       cos3theta = cos(3*theta0)
       sin4theta = sin(4*theta0)
       cos4theta = cos(4*theta0)

       Newton_residual(1) = Newton_residual(1) + r0 * r0 * r0 * (g3s1 * sintheta + g3s3 * sin3theta + g3c1 * costheta + g3c3 * cos3theta) \
            + r0 * r0 * r0 * r0 * (g40 + g4s2 * sin2theta + g4s4 * sin4theta + g4c2 * cos2theta + g4c4 * cos4theta)

       Newton_residual(2) = Newton_residual(2) + r0 * r0 * r0 * (g3s1 * costheta + g3s3 * 3 * cos3theta + g3c1 * (-sintheta) + g3c3 * (-3*sin3theta)) \
            + r0 * r0 * r0 * r0 * (g4s2 * 2 * cos2theta + g4s4 * 4 * cos4theta + g4c2 * (-2*sin2theta) + g4c4 * (-4*sin4theta))
    end if

    Newton_residual_sqnorm = Newton_residual(1) * Newton_residual(1) + Newton_residual(2) * Newton_residual(2)

  end subroutine r_singularity_residual

  subroutine r_singularity_Jacobian

    implicit none

    real(dp) :: inv_determinant, Jacobian(2,2)
    real(dp) :: theta0, r0

    r0 = state(1)
    theta0 = state(2)

    ! If ghat = sqrt{g}/r,
    ! Jacobian = [d ghat / d r,           d ghat / d theta    ]
    !            [d^2 ghat / d r d theta, d^2 ghat / d theta^2]
    Jacobian(1,1) = g1c * costheta + 2 * r0 * (g20 + g2s * sin2theta + g2c * cos2theta)
    Jacobian(1,2) = r0 * (-g1c * sintheta) + 2 * r0 * r0 * (g2s * cos2theta - g2c * sin2theta)
    Jacobian(2,1) = -g1c * sintheta + 4 * r0 * (g2s * cos2theta - g2c * sin2theta)
    Jacobian(2,2) = -r0 * (g1c * costheta) - 4 * r0 * r0 * (g2s * sin2theta + g2c * cos2theta)

    if (r_singularity_high_order) then
       ! d ghat / d r
       Jacobian(1,1) = Jacobian(1,1) + 3 * r0 * r0 * (g3s1 * sintheta + g3s3 * sin3theta + g3c1 * costheta + g3c3 * cos3theta) \
            + 4 * r0 * r0 * r0 * (g40 + g4s2 * sin2theta + g4s4 * sin4theta + g4c2 * cos2theta + g4c4 * cos4theta)

       ! d ghat / d theta
       Jacobian(1,2) = Jacobian(1,2) + r0 * r0 * r0 * (g3s1 * costheta + g3s3 * 3 * cos3theta + g3c1 * (-sintheta) + g3c3 * (-3*sin3theta)) \
            + r0 * r0 * r0 * r0 * (g4s2 * 2 * cos2theta + g4s4 * 4 * cos4theta + g4c2 * (-2*sin2theta) + g4c4 * (-4*sin4theta))

       ! d^2 ghat / d r d theta
       Jacobian(2,1) = Jacobian(2,1) + 3 * r0 * r0 * (g3s1 * costheta + g3s3 * 3 * cos3theta + g3c1 * (-sintheta) + g3c3 * (-3*sin3theta)) \
            + 4 * r0 * r0 * r0 * (g4s2 * 2 * cos2theta + g4s4 * 4 * cos4theta + g4c2 * (-2*sin2theta) + g4c4 * (-4*sin4theta))

       ! d^2 ghat / d theta^2
       Jacobian(2,2) = Jacobian(2,2) - r0 * r0 * r0 * (g3s1 * sintheta + g3s3 * 9 * sin3theta + g3c1 * costheta + g3c3 * 9 * cos3theta) \
            - r0 * r0 * r0 * r0 * (g4s2 * 4 * sin2theta + g4s4 * 16 * sin4theta + g4c2 * 4 * cos2theta + g4c4 * 16 * cos4theta)
       
    end if

    !if (j==80) then
    !   print *," ZZZ Jacobian:"
    !   print "(2(es24.15))", Jacobian(1,:)
    !   print "(2(es24.15))", Jacobian(2,:)
    !end if

    inv_determinant = 1 / (Jacobian(1,1) * Jacobian(2,2) - Jacobian(1,2) * Jacobian(2,1))
    ! Inverse of a 2x2 matrix:
    inv_Jacobian(1,1) =  Jacobian(2,2) * inv_determinant
    inv_Jacobian(1,2) = -Jacobian(1,2) * inv_determinant
    inv_Jacobian(2,1) = -Jacobian(2,1) * inv_determinant
    inv_Jacobian(2,2) =  Jacobian(1,1) * inv_determinant

  end subroutine r_singularity_Jacobian

end subroutine quasisymmetry_max_r_before_singularity

! ---------------------------------------------------

!> Compute the roots of a quartic polynomial by computing eigenvalues of a companion matrix
!>
!> This subroutine follows the same algorithm as Matlab's 'roots' routine.
!> @params coefficients Ordered the same way as in matlab, from the coefficient of x^4 to the coefficient of x^0.
!> @params real_parts Real parts of the roots
!> @params imag_parts Imaginary parts of the roots
subroutine quasisymmetry_quartic_roots(coefficients, real_parts, imag_parts)

  use quasisymmetry_variables, only: dp

  implicit none
  real(dp), intent(in) :: coefficients(5)
  real(dp), intent(out) :: real_parts(4), imag_parts(4)

  real(dp) :: matrix(4,4)
  integer, parameter :: LWORK = 100 ! Work array for LAPACK
  real(dp) :: WORK(LWORK), VL(1), VR(1)
  integer :: INFO

  matrix(1,1) = -coefficients(2) / coefficients(1)
  matrix(1,2) = -coefficients(3) / coefficients(1)
  matrix(1,3) = -coefficients(4) / coefficients(1)
  matrix(1,4) = -coefficients(5) / coefficients(1)

  matrix(2:4,:) = 0.0_dp

  matrix(2,1) = 1.0_dp
  matrix(3,2) = 1.0_dp
  matrix(4,3) = 1.0_dp

  call dgeev('N', 'N', 4, matrix, 4, real_parts, imag_parts, VL, 1, VR, 1, WORK, LWORK, INFO)
  if (INFO .ne. 0) then
     print *,"Error in DGEEV: info=",INFO
     stop
  end if

end subroutine quasisymmetry_quartic_roots
"""



#!/usr/bin/env python3

"""
This module contains a subroutine for making spectral differentiation matrices.
"""

import numpy as np
from scipy.linalg import toeplitz

def spectral_diff_matrix(n, xmin=0, xmax=2*np.pi):
    """
    Return the spectral differentiation matrix for n grid points
    on the periodic domain [xmax, xmax). This routine is based on the
    matlab code in the DMSuite package by S.C. Reddy and J.A.C. Weideman, available at
    http://www.mathworks.com/matlabcentral/fileexchange/29
    or here:
    http://dip.sun.ac.za/~weideman/research/differ.html  
    """

    h = 2 * np.pi / n
    kk = np.arange(1, n)
    n1 = int(np.floor((n - 1) / 2))
    n2 = int(np.ceil((n - 1) / 2))
    if np.mod(n, 2) == 0:
        topc = 1 / np.tan(np.arange(1, n2 + 1) * h / 2)
        temp = np.concatenate((topc, -np.flip(topc[0:n1])))
    else:
        topc = 1 / np.sin(np.arange(1, n2 + 1) * h / 2)
        temp = np.concatenate((topc, np.flip(topc[0:n1])))

    col1 = np.concatenate(([0], 0.5 * ((-1) ** kk) * temp))
    row1 = -col1
    D = 2 * np.pi / (xmax - xmin) * toeplitz(col1, r=row1)
    return D



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="qicna",
    version="0.3.2",
    author="Rogerio Jorge and Matt Landreman",
    author_email="rogerio.jorge@tecnico.ulisboa.pt",
    description="Quasi-isodynamic Stellarator Construction",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    url="https://github.com/rogeriojorge/pyQIC",
    #install_requires=['numpy', 'scipy'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
#    packages=["qicna"],



"""
This module contains a function to plot a near-axis surface.
"""

import numpy as np
from scipy.interpolate import interp2d, interp1d
import matplotlib.pyplot as plt
from matplotlib import cm
import matplotlib.colors as clr
from matplotlib.colors import LightSource
import matplotlib.ticker as tck
from .util import to_Fourier

def plot(self, newfigure=True, show=True, savefig=None):
    """
    Generate a matplotlib figure with an array of plots, showing the
    toroidally varying properties of the configuration.

    Args:
        newfigure: Whether to create a new matplotlib figure.
        show: Whether to call matplotlib's ``show()`` function after making the plots.
    """
    if newfigure:
        f = plt.figure(figsize=(14, 7))
    plt.rcParams.update({'font.size': 6})
    if self.order == 'r1':
        if self.omn:
            nrows = 4
            ncols = 6
        else:
            nrows = 3
            ncols = 6
    elif self.order == 'r2':
        nrows = 4
        ncols = 8
        if self.omn:
            nrows=6
    elif self.order == 'r3':
        nrows = 6
        ncols = 8
    else:
        raise RuntimeError('Should not get here')
    jplot = 1

    def subplot(title, data=None, y0=False):
        """
        data is assumed to correspond to title, unless specified otherwise.
        Set y0 to True to avoid suppressed 0 on the y axis.
        """
        nonlocal jplot
        if data is None:
            data = eval('self.' + title)
        plt.subplot(nrows, ncols, jplot)
        jplot = jplot + 1
        plt.plot(self.phi, data, label=title)
        plt.xlabel(r'$\phi$')
        plt.title(title)
        if y0:
            plt.ylim(bottom=0)
        plt.xlim((0, self.phi[-1]))

    subplot('R0')
    subplot('Z0')
    subplot('R0p')
    subplot('Z0p')
    subplot('R0pp')
    subplot('Z0pp')
    subplot('R0ppp')
    subplot('Z0ppp')
    subplot('curvature')
    subplot('torsion')
    subplot('sigma')
    subplot('X1s')
    subplot('X1c')
    subplot('Y1c')
    subplot('Y1s')
    subplot('elongation', y0=True)
    subplot('L_grad_B', y0=True)
    subplot('1/L_grad_B', data=self.inv_L_grad_B)
    # if self.omn:
    #     subplot('alpha')
    if self.order != 'r1':
        jplot -= 2
        subplot('L_grad_grad_B')
        plt.title('scale lengths')
        plt.legend(loc=0, fontsize=5)
        plt.ylim(0, max(max(self.L_grad_B), max(self.L_grad_grad_B)))
        
        subplot('1/L_grad_grad_B', self.grad_grad_B_inverse_scale_length_vs_varphi)
        plt.title('inv scale lengths')
        plt.legend(loc=0, fontsize=5)
        plt.ylim(0, max(max(self.inv_L_grad_B), max(self.grad_grad_B_inverse_scale_length_vs_varphi)))
        
        subplot('V1')
        subplot('V2')
        subplot('V3')
        subplot('X20')
        subplot('X2c')
        subplot('X2s')
        subplot('Y20')
        subplot('Y2c')
        subplot('Y2s')
        subplot('Z20')
        subplot('Z2c')
        subplot('Z2s')
        data = self.r_singularity_vs_varphi
        data[data > 1e20] = np.NAN
        subplot('r_singularity', data=data, y0=True)
    if self.omn:
        subplot('alpha')
        jplot -= 1
        subplot('alpha_no_buffer')
        plt.title('alpha')
        plt.legend(loc=0, fontsize=5)

        plt.subplot(nrows, ncols, jplot)
        d_alpha_iota_d_varphi_der    = [self.d_alpha_iota_d_varphi]
        d_alpha_notIota_d_varphi_der = [self.d_alpha_notIota_d_varphi]
        d_alpha_der = [d_alpha_iota_d_varphi_der[0] * self.iota + d_alpha_notIota_d_varphi_der[0]]
        ders = [1]
        plt.plot(d_alpha_der[0], label='n=1')
        plt.xlabel(r'$\phi$')
        plt.title(r'$\alpha^{(n)}$')
        self.d_alpha_der_diff = [d_alpha_der[0][-1]-d_alpha_der[0][0]]
        for n in range(2,5):
            d_alpha_iota_d_varphi_der.append(np.matmul(self.d_d_varphi,d_alpha_iota_d_varphi_der[n-2]))
            d_alpha_notIota_d_varphi_der.append(np.matmul(self.d_d_varphi,d_alpha_notIota_d_varphi_der[n-2]))
            d_alpha_der.append(d_alpha_iota_d_varphi_der[n-1] * self.iota + d_alpha_notIota_d_varphi_der[n-1])
            self.d_alpha_der_diff.append(d_alpha_der[n-1][-1]-d_alpha_der[n-1][0])
            ders.append(n)
            plt.plot(d_alpha_der[n-1], label='n='+str(n))
        plt.legend()
        jplot += 1

        plt.subplot(nrows, ncols, jplot)
        plt.plot(ders,self.d_alpha_der_diff)
        plt.xlabel(r'$n=$Order of the derivative')
        plt.title(r'$\alpha^{(n)}(2\pi)-\alpha^{(n)}(0)$')
        jplot += 1
    if self.order != 'r1':
        subplot('B20')
        if self.omn:
            # jplot -= 1
            # subplot('B2QI_exact')
            # plt.legend(loc=0, fontsize=5)
            subplot('B2cQI')
            subplot('B2sQI')
            # jplot -= 1
            # subplot('B2QI_exact')
            # plt.legend(loc=0, fontsize=5)
            subplot('B20QI_deviation')
            subplot('B2cQI_deviation')
            subplot('B2sQI_deviation')
            subplot('B2c_array')
            subplot('B2s_array')
    if self.omn:
        subplot('d')
        subplot('gamma')
    if self.order != 'r1':
        if self.order != 'r2':
            subplot('X3c1')
            subplot('X3s1')
            subplot('Y3c1')
            subplot('Y3s1')

    plt.tight_layout()
    if savefig!=None:
        plt.savefig(savefig+'.pdf')
    if show:
        plt.show()

def set_axes_equal(ax):
    '''
    Make axes of 3D plot have equal scale so that spheres appear as spheres,
    cubes as cubes, etc..  This is one possible solution to Matplotlib's
    ax.set_aspect('equal') and ax.axis('equal') not working for 3D.

    Args:
      ax: a matplotlib axis, e.g., as output from plt.gca().
    '''
    x_limits = ax.get_xlim3d()
    y_limits = ax.get_ylim3d()
    z_limits = ax.get_zlim3d()

    x_range = abs(x_limits[1] - x_limits[0])
    x_middle = np.mean(x_limits)
    y_range = abs(y_limits[1] - y_limits[0])
    y_middle = np.mean(y_limits)
    z_range = abs(z_limits[1] - z_limits[0])
    z_middle = np.mean(z_limits)

    # The plot bounding box is a sphere in the sense of the infinity
    # norm, hence call half the max range the plot radius.
    plot_radius = 0.5 * max([x_range, y_range, z_range])

    ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
    ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
    ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])

def create_subplot(ax, x_2D_plot, y_2D_plot, z_2D_plot, colormap, elev=90, azim=45, dist=7, **kwargs):
    '''
    Construct the surface given a surface in cartesian coordinates
    x_2D_plot, y_2D_plot, z_2D_plot already with phi=[0,2*pi].
    A matplotlib figure with elements fig, ax
    must have been previously created.

    Args:
        ax: matplotlib figure instance
        x_2d_plot: 2D array for the x coordinates of the surface
        y_2d_plot: 2D array for the x coordinates of the surface
        z_2d_plot: 2D array for the x coordinates of the surface
        elev: elevation angle for the camera view
        azim: azim angle for the camera view
        distance: distance parameter for the camera view
    '''
    ax.plot_surface(x_2D_plot, y_2D_plot, z_2D_plot, facecolors=colormap,
                    rstride=1, cstride=1, antialiased=False,
                    linewidth=0, alpha=1., shade=False, **kwargs)
    set_axes_equal(ax)
    ax.set_axis_off()
    ax.dist = dist
    ax.elev = elev
    ax.azim = azim

def create_field_lines(qic, alphas, X_2D, Y_2D, Z_2D, phimax=2*np.pi, nphi=500):
    '''
    Function to compute the (X, Y, Z) coordinates of field lines at
    several alphas, where alpha = theta-iota*varphi with (theta,varphi)
    the Boozer toroidal angles. This function relies on a 2D interpolator
    from the scipy library to smooth out the lines

    Args:
      qic: instance of self
      alphas: array of field line labels alpha
      X_2D: 2D array for the x components of the surface
      Y_2D: 2D array for the y components of the surface
      Z_2D: 2D array for the z components of the surface
      phimax: maximum value for the field line following angle phi
      nphi: grid resolution for the output fieldline
    '''
    phi_array = np.linspace(0,phimax,nphi,endpoint=False)
    fieldline_X = np.zeros((len(alphas),nphi))
    fieldline_Y = np.zeros((len(alphas),nphi))
    fieldline_Z = np.zeros((len(alphas),nphi))
    [ntheta_RZ,nphi_RZ] = X_2D.shape
    phi1D   = np.linspace(0,2*np.pi,nphi_RZ)
    theta1D = np.linspace(0,2*np.pi,ntheta_RZ)
    X_2D_spline = interp2d(phi1D, theta1D, X_2D, kind='cubic')
    Y_2D_spline = interp2d(phi1D, theta1D, Y_2D, kind='cubic')
    Z_2D_spline = interp2d(phi1D, theta1D, Z_2D, kind='cubic')
    for i in range(len(alphas)):
        for j in range(len(phi_array)):
            phi_mod = np.mod(phi_array[j],2*np.pi)
            varphi0=qic.nu_spline(phi_array[j])+2*phi_array[j]-phi_mod
            theta_fieldline=qic.iota*varphi0+alphas[i]
            theta_fieldline_mod=np.mod(theta_fieldline,2*np.pi)
            fieldline_X[i,j] = X_2D_spline(phi_mod,theta_fieldline_mod)[0]
            fieldline_Y[i,j] = Y_2D_spline(phi_mod,theta_fieldline_mod)[0]
            fieldline_Z[i,j] = Z_2D_spline(phi_mod,theta_fieldline_mod)[0]
    return fieldline_X, fieldline_Y, fieldline_Z

def create_subplot_mayavi(mlab, R, alphas, x_2D_plot, y_2D_plot, z_2D_plot,
                          fieldline_X, fieldline_Y, fieldline_Z,
                          Bmag, degrees_array_x, degrees_array_z, shift_array):
    '''
    Plotting routine for a mayavi figure instance that plots
    both the surface and the field lines together. The number
    of surfaces to plot is specified by the length of the
    array degrees_array_x

    Args:
      mlab: mayavi package
      R: scipy rotation vector package
      alphas: array of field line labels alpha
      x_2D_plot: 2D array for the x components of the surface
      y_2D_plot: 2D array for the y components of the surface
      z_2D_plot: 2D array for the z components of the surface
      fieldline_X: 2D array for the x components of the field line
      fieldline_Y: 2D array for the x components of the field line
      fieldline_Z: 2D array for the x components of the field line
      Bmag: 2D array for the magnetic field modulus on the (theta,phi) meshgrid
      degrees_array_x: 1D array with the rotation angles in the x direction for each surface
      degrees_array_z: 1D array with the rotation angles in the z direction for each surface
      shift_array: 1D array with a shift in the y direction for each surface
    '''
    assert len(degrees_array_x) == len(degrees_array_z) == len(shift_array)
    for i in range(len(degrees_array_x)):
        # The surfaces and field lines are rotated first in the
        # z direction and then in the x direction
        rx= R.from_euler('x', degrees_array_x[i], degrees=True)
        rz= R.from_euler('z', degrees_array_z[i], degrees=True)
        # Initialize rotated arrays
        x_2D_plot_rotated = np.zeros((x_2D_plot.shape[0],x_2D_plot.shape[1]))
        y_2D_plot_rotated = np.zeros((x_2D_plot.shape[0],x_2D_plot.shape[1]))
        z_2D_plot_rotated = np.zeros((x_2D_plot.shape[0],x_2D_plot.shape[1]))
        fieldline_X_rotated = np.zeros((fieldline_X.shape[0],fieldline_X.shape[1]))
        fieldline_Y_rotated = np.zeros((fieldline_X.shape[0],fieldline_X.shape[1]))
        fieldline_Z_rotated = np.zeros((fieldline_X.shape[0],fieldline_X.shape[1]))
        # Rotate surfaces
        for th in range(x_2D_plot.shape[0]):
            for ph in range(x_2D_plot.shape[1]):
                [x_2D_plot_rotated[th,ph], y_2D_plot_rotated[th,ph], z_2D_plot_rotated[th,ph]] = rx.apply(rz.apply(np.array([x_2D_plot[th,ph], y_2D_plot[th,ph], z_2D_plot[th,ph]])))
        # Rotate field lines
        for th in range(fieldline_X.shape[0]):
            for ph in range(fieldline_X.shape[1]):
                [fieldline_X_rotated[th,ph], fieldline_Y_rotated[th,ph], fieldline_Z_rotated[th,ph]] = rx.apply(rz.apply(np.array([fieldline_X[th,ph], fieldline_Y[th,ph], fieldline_Z[th,ph]])))
        # Plot surfaces
        mlab.mesh(x_2D_plot_rotated, y_2D_plot_rotated-shift_array[i], z_2D_plot_rotated, scalars=Bmag, colormap='viridis', opacity=1.0)
        # Plot field lines
        for j in range(len(alphas)):
            mlab.plot3d(fieldline_X_rotated[j], fieldline_Y_rotated[j]-shift_array[i], fieldline_Z_rotated[j], color=(0,0,0), line_width=0.002, tube_radius=0.008)

def get_boundary(self, r=0.1, ntheta=40, nphi=130, ntheta_fourier=20, mpol=13, ntor=25):
    '''
    Function that, for a given near-axis radial coordinate r, outputs
    the [X,Y,Z,R] components of the boundary. The resolution along the toroidal
    angle phi is equal to the resolution nphi for the axis, while ntheta
    is specified by the used.

    Args:
      r (float): near-axis radius r where to create the surface
      ntheta (int): Number of grid points to plot in the poloidal angle.
      nphi   (int): Number of grid points to plot in the toroidal angle.
      ntheta_fourier (int): Resolution in the Fourier transform to cylindrical coordinates
      mpol: resolution in poloidal Fourier space
      ntor: resolution in toroidal Fourier space
    '''
    # Get surface shape at fixed off-axis toroidal angle phi
    R_2D, Z_2D, _ = self.Frenet_to_cylindrical(r, ntheta=ntheta_fourier)
    # Get Fourier coefficients in order to plot with arbitrary resolution
    RBC, RBS, ZBC, ZBS = to_Fourier(R_2D, Z_2D, self.nfp, mpol=mpol, ntor=ntor, lasym=self.lasym)
    if not self.lasym:
        RBS = np.zeros((int(2*ntor+1),int(mpol+1)))
        ZBC = np.zeros((int(2*ntor+1),int(mpol+1)))

    theta1D = np.linspace(0, 2*np.pi, ntheta)
    phi1D = np.linspace(0, 2*np.pi, nphi)
    phi2D, theta2D = np.meshgrid(phi1D, theta1D)
    R_2Dnew = np.zeros((ntheta, nphi))
    Z_2Dnew = np.zeros((ntheta, nphi))
    for m in range(mpol + 1):
        for n in range(-ntor, ntor + 1):
            angle = m * theta2D - n * self.nfp * phi2D
            R_2Dnew += RBC[n+ntor,m] * np.cos(angle) + RBS[n+ntor,m] * np.sin(angle)
            Z_2Dnew += ZBC[n+ntor,m] * np.cos(angle) + ZBS[n+ntor,m] * np.sin(angle)

    # X, Y, Z arrays for the whole surface
    x_2D_plot = R_2Dnew * np.cos(phi1D)
    y_2D_plot = R_2Dnew * np.sin(phi1D)
    z_2D_plot = Z_2Dnew

    return x_2D_plot, y_2D_plot, z_2D_plot, R_2Dnew

def plot_boundary(self, r=0.1, ntheta=80, nphi=150, ntheta_fourier=20, nsections=8,
         fieldlines=False, savefig=None, colormap=None, azim_default=None, threeD=True, n_field_lines=1,
         show=True, **kwargs):
    """
    Plot the boundary of the near-axis configuration. There are two main ways of
    running this function.

    If ``fieldlines=False`` (default), 2 matplotlib figures are generated:

        - A 2D plot with several poloidal planes at the specified radius r with the
          corresponding location of the magnetic axis.

        - A 3D plot with the flux surface and the magnetic field strength
          on the surface.

    If ``fieldlines=True``, both matplotlib and mayavi are required, and
    the following 2 figures are generated:

        - A 2D matplotlib plot with several poloidal planes at the specified radius r with the
          corresponding location of the magnetic axis.

        - A 3D mayavi figure with the flux surface the magnetic field strength
          on the surface and several magnetic field lines.

    Args:
      r (float): near-axis radius r where to create the surface
      ntheta (int): Number of grid points to plot in the poloidal angle.
      nphi   (int): Number of grid points to plot in the toroidal angle.
      ntheta_fourier (int): Resolution in the Fourier transform to cylindrical coordinates
      nsections (int): Number of poloidal planes to show.
      fieldlines (bool): Specify if fieldlines are shown. Using mayavi instead of matplotlib due to known bug https://matplotlib.org/2.2.2/mpl_toolkits/mplot3d/faq.html
      savefig (str): Filename prefix for the png files to save.
        Note that a suffix including ``.png`` will be appended.
        If ``None``, no figure files will be saved.
      colormap (cmap): Custom colormap for the 3D plots
      azim_default: Default azimuthal angle for the three subplots in the 3D surface plot
      show: Whether or not to call the matplotlib/mayavi ``show()`` command.
      kwargs: Any additional key-value pairs to pass to matplotlib's plot_surface.

    This function generates plots similar to the ones below:

    .. image:: 3dplot1.png
       :width: 200

    .. image:: 3dplot2.png
       :width: 200

    .. image:: poloidalplot.png
       :width: 200
    """
    x_2D_plot, y_2D_plot, z_2D_plot, R_2D_plot = self.get_boundary(r=r, ntheta=ntheta, nphi=nphi, ntheta_fourier=ntheta_fourier)
    phi = np.linspace(0, 2 * np.pi, nphi)  # Endpoint = true and no nfp factor, because this is what is used in get_boundary()
    R_2D_spline = interp1d(phi, R_2D_plot, axis=1)
    z_2D_spline = interp1d(phi, z_2D_plot, axis=1)
    ## Poloidal plot
    phi1dplot_RZ = np.linspace(0, 2 * np.pi / self.nfp, nsections, endpoint=False)
    fig_poloidal = plt.figure(figsize=(7, 5), dpi=80)
    ax  = plt.gca()
    for i, phi in enumerate(phi1dplot_RZ):
        phinorm = phi * self.nfp / (2 * np.pi)
        if phinorm == 0:
            label = r'$\phi$=0'
        elif phinorm == 0.125:
            label = r'$\phi={\pi}/$' + str(4 * self.nfp)
        elif phinorm == 0.25:
            label = r'$\phi={\pi}/$' + str(2 * self.nfp)
        elif phinorm == 0.375:
            label = r'$\phi={3\pi}/$' + str(4 * self.nfp)
        elif phinorm == 0.5:
            label = r'$\phi=\pi/$' + str(self.nfp)
        elif phinorm == 0.625:
            label = r'$\phi={5\pi}/$' + str(4 * self.nfp)
        elif phinorm == 0.75:
            label = r'$\phi={3\pi}/$' + str(2 * self.nfp)
        elif phinorm == 0.875:
            label = r'$\phi={7\pi}/$' + str(4 * self.nfp)
        else:
            label = '_nolegend_'
        color = next(ax._get_lines.prop_cycler)['color']
        # Plot location of the axis
        plt.plot(self.R0_func(phi), self.Z0_func(phi), marker="x", linewidth=2, label=label, color=color)
        # Plot poloidal cross-section
        plt.plot(R_2D_spline(phi), z_2D_spline(phi), color=color)
    plt.xlabel('R (meters)', fontsize=14)
    plt.ylabel('Z (meters)', fontsize=14)
    ax.tick_params(axis='both', which='major', labelsize=12)
    ax.tick_params(axis='both', which='minor', labelsize=12)
    plt.legend(loc=2, prop={'size': 8})
    plt.tight_layout()
    ax.set_aspect('equal')
    if savefig != None:
        fig_poloidal.savefig(savefig + '_poloidal.pdf')

    ## 3D plot
    # Set the default azimuthal angle of view in the 3D plot
    # QH stellarators look rotated in the phi direction when
    # azim_default = 0
    if threeD==True:
        if azim_default == None:
            if self.omn == True:
                azim_default = -90
            else:
                if self.helicity == 0:
                    azim_default = 0
                else:
                    azim_default = 45
            # Define the magnetic field modulus and create its theta,phi array
            # The norm instance will be used as the colormap for the surface
            theta1D = np.linspace(0, 2 * np.pi, ntheta)
            phi1D = np.linspace(0, 2 * np.pi, nphi)
            phi2D, theta2D = np.meshgrid(phi1D, theta1D)
            # Create a color map similar to viridis 
            Bmag = self.B_mag(r, theta2D, phi2D)
            norm = clr.Normalize(vmin=Bmag.min(), vmax=Bmag.max())
            if fieldlines==False:
                if colormap==None:
                    # Cmap similar to quasisymmetry papers
                    # cmap = clr.LinearSegmentedColormap.from_list('qs_papers',['#4423bb','#4940f4','#2e6dff','#0097f2','#00bacc','#00cb93','#00cb93','#7ccd30','#fbdc00','#f9fc00'], N=256)
                    cmap = cm.viridis
                    # Add a light source so the surface looks brighter
                    ls = LightSource(azdeg=0, altdeg=10)
                    cmap_plot = ls.shade(Bmag, cmap, norm=norm)
                # Create the 3D figure and choose the following parameters:
                # gsParams: extension in the top, bottom, left right directions for each subplot
                # elevParams: elevation (distance to the plot) for each subplot
                fig = plt.figure(constrained_layout=False, figsize=(4.5, 8))
                gsParams = [[1.02,-0.3,0.,0.85], [1.09,-0.3,0.,0.85], [1.12,-0.15,0.,0.85]]
                elevParams = [90, 30, 5]
                for i in range(len(gsParams)):
                    gs = fig.add_gridspec(nrows=3, ncols=1,
                                        top=gsParams[i][0], bottom=gsParams[i][1],
                                        left=gsParams[i][2], right=gsParams[i][3],
                                        hspace=0.0, wspace=0.0)
                    ax = fig.add_subplot(gs[i, 0], projection='3d')
                    create_subplot(ax, x_2D_plot, y_2D_plot, z_2D_plot, cmap_plot, elev=elevParams[i], azim=azim_default, **kwargs)
                # Create color bar with axis placed on the right
                cbar_ax = fig.add_axes([0.85, 0.2, 0.03, 0.6])
                m = cm.ScalarMappable(cmap=cmap, norm=norm)
                m.set_array([])
                cbar = plt.colorbar(m, cax=cbar_ax)
                cbar.ax.set_title(r'$|B| [T]$')
                # Save figure
                if savefig != None:
                    fig.savefig(savefig + '3D.png')
                if show:
                    # Show figures
                    plt.show()
            else:
                ## X, Y, Z arrays for the field lines
                # Plot different field lines corresponding to different alphas
                # where alpha=theta-iota*varphi with (theta,varphi) the Boozer angles
                #alphas = [0, np.pi/4, np.pi/2, 3*np.pi/4, np.pi, 5*np.pi/4, 3*np.pi/2, 7*np.pi/4]
                alphas = np.linspace(0.7, 2 * np.pi + 0.7, n_field_lines, endpoint=False)
                # Create the field line arrays
                fieldline_X, fieldline_Y, fieldline_Z = create_field_lines(self, alphas, x_2D_plot, y_2D_plot, z_2D_plot)
                # Define the rotation arrays for the subplots
                degrees_array_x = [0., 81.]#[0., -66., 81.] # degrees for rotation in x
                degrees_array_z = [azim_default, azim_default]#[azim_default, azim_default, azim_default] # degrees for rotation in z
                shift_array   = [-0.8, 1.0]#[-0.9, 0.6, 1.8]
                # Import mayavi and rotation packages (takes a few seconds)
                from mayavi import mlab
                from scipy.spatial.transform import Rotation as R
                if show:
                    # Show RZ plot
                    plt.show()
                # Create 3D figure
                fig_3d = mlab.figure(bgcolor=(1,1,1), size=(580,600))
                # Create subplots
                create_subplot_mayavi(mlab, R, alphas, x_2D_plot, y_2D_plot, z_2D_plot,
                                    fieldline_X, fieldline_Y, fieldline_Z,
                                    Bmag, degrees_array_x, degrees_array_z, shift_array)
                # Create a good camera angle
                mlab.view(azimuth=0, elevation=0, distance=8.0, focalpoint=(0,0,0), figure=fig_3d)
                # Create the colorbar and change its properties
                cb = mlab.colorbar(orientation='vertical', title='|B| [T]', nb_labels=7)
                cb.scalar_bar.unconstrained_font_size = True
                cb.label_text_property.font_family = 'times'
                cb.label_text_property.bold = 0
                cb.label_text_property.font_size=20
                cb.label_text_property.color=(0,0,0)
                cb.title_text_property.font_family = 'times'
                cb.title_text_property.font_size=20
                cb.title_text_property.color=(0,0,0)
                cb.title_text_property.bold = 1
                # Save figure
                if savefig != None:
                    mlab.savefig(filename=savefig+'3D_fieldlines.png', figure=fig_3d)
                if show:
                    # Show mayavi plot
                    mlab.show()
                    # Close mayavi plots
                    mlab.close(all=True)
    else:
        plt.show()

def B_fieldline(self, r=0.1, alpha=0, phimax=None, nphi=400, show=True, savefig=None):
    '''
    Plot the modulus of the magnetic field B along a field line with
    the Boozer toroidal angle varphi acting as a field-line following
    coordinate

    Args:
      r (float): near-axis radius r where to create the surface
      alpha (float): Field-line label
      phimax (float): Maximum value of the field-line following parameter varphi
      nphi (int): resolution of the phi grid
      show (bool): Whether or not to call the matplotlib ``show()`` command.
    '''
    if phimax is None:
        phimax = 30 * np.pi / abs(self.iota)
    varphi_array = np.linspace(0, phimax, nphi)
    fig_B_fieldline, ax = plt.subplots(1, 1, figsize=(10, 6), dpi=80, facecolor='w', edgecolor='k')
    plt.xlabel(r'$\varphi$')
    plt.ylabel(r'$B(\varphi)$')
    plt.title("r = " + str(r) + ", alpha = " + str(alpha))
    theta = alpha + self.iota * varphi_array
    plt.plot(varphi_array/np.pi, self.B_mag(r, theta, varphi_array, Boozer_toroidal=True))
    ax.xaxis.set_major_formatter(tck.FormatStrFormatter('%g $\pi$'))
    ax.xaxis.set_major_locator(tck.MultipleLocator(base=2))
    plt.tight_layout()
    if savefig != None:
        fig_B_fieldline.savefig(savefig + '_B_fieldline.pdf')
    if show:
        plt.show()

def B_contour(self, r=0.1, ntheta=100, nphi=120, ncontours=20, B0=1, show=True, savefig=None):
    '''
    Plot contours of constant B, with B the modulus of the
    magnetic field, as a function of Boozer coordinates theta and varphi

    Args:
      r (float): near-axis radius r where to create the surface
      ntheta (int): Number of grid points to plot in the Boozer poloidal angle.
      nphi   (int): Number of grid points to plot in the Boozer toroidal angle.
      ncontours (int): number of contours to show in the plot
      show (bool): Whether or not to call the matplotlib ``show()`` command.
    '''
    theta_array=np.linspace(0,2*np.pi,ntheta)
    phi_array=np.linspace(0,2*np.pi/self.nfp,nphi)
    theta_2D, phi_2D = np.meshgrid(theta_array,phi_array)
    magB_2D = self.B_mag(r,theta_2D,phi_2D,Boozer_toroidal=True,B0=B0)
    fig_B_contour, ax=plt.subplots(1,1)
    contourplot = ax.contour(phi_2D/np.pi, theta_2D/np.pi, magB_2D, ncontours, cmap=cm.plasma, linewidths=2.0)
    fig_B_contour.colorbar(contourplot)
    # ax.set_title('|B| for r=' + str(r))
    ax.set_xlabel(r'Boozer $\varphi$', fontsize=16)
    ax.set_ylabel(r'Boozer $\theta$', fontsize=16)
    ax.xaxis.set_major_formatter(tck.FormatStrFormatter('%g $\pi$'))
    ax.yaxis.set_major_formatter(tck.FormatStrFormatter('%g $\pi$'))
    ax.xaxis.set_major_locator(tck.MultipleLocator(base=0.5/self.nfp))
    ax.yaxis.set_major_locator(tck.MultipleLocator(base=0.5))
    plt.tight_layout()
    if savefig != None:
        fig_B_contour.savefig(savefig + '_B_contour.pdf')
    if show:
        plt.show()

def B_densityplot(self, r=0.1, ntheta=250, nphi=250, B0=1, show=True, savefig=None):
    '''
    Density plot of B, with B the modulus of the
    magnetic field, as a function of Boozer coordinates theta and varphi

    Args:
      r (float): near-axis radius r where to create the surface
      ntheta (int): Number of grid points to plot in the Boozer poloidal angle.
      nphi   (int): Number of grid points to plot in the Boozer toroidal angle.
      show (bool): Whether or not to call the matplotlib ``show()`` command.
    '''
    theta_array=np.linspace(0,2*np.pi,ntheta)
    phi_array=np.linspace(0,2*np.pi/self.nfp,nphi)
    theta_2D, phi_2D = np.meshgrid(theta_array,phi_array)
    magB_2D = self.B_mag(r,theta_2D,phi_2D,Boozer_toroidal=True,B0=B0)
    # contourplot = ax.contour(phi_2D/np.pi, theta_2D/np.pi, magB_2D, ncontours, cmap=cm.plasma, linewidths=2.0)
    fig_B_densityploy, ax=plt.subplots(1,1)
    contourplot = ax.imshow(magB_2D.transpose(), extent=[0, 2/self.nfp, 0, 2], cmap=cm.plasma, aspect='auto')
    fig_B_densityploy.colorbar(contourplot)
    ax.set_title('|B| for r=' + str(r))
    ax.set_xlabel(r'$\varphi$')
    ax.set_ylabel(r'$\theta$')
    ax.xaxis.set_major_formatter(tck.FormatStrFormatter('%g $\pi$'))
    ax.yaxis.set_major_formatter(tck.FormatStrFormatter('%g $\pi$'))
    ax.xaxis.set_major_locator(tck.MultipleLocator(base=0.5/self.nfp))
    ax.yaxis.set_major_locator(tck.MultipleLocator(base=0.5))
    plt.tight_layout()
    if savefig != None:
        fig_B_densityploy.savefig(savefig + '_B_densityploy.pdf')
    if show:
        plt.show()

def plot_axis(self, nphi=100, frenet=True, nphi_frenet=80, frenet_factor=0.12, savefig=None, show=True):
    '''
    Plot axis shape and the Frenet-Serret frame along
    the axis (optional). If frenet is true, creates
    a mayavi instance showing the axis and nphi_frenet
    times 3 vectors, corresponding to the tangent, normal and
    binormal vectors. If frenet is false, creates a
    matplotlib instance with only a single axis shape
    curve shown.

    Args:
      nphi (int): Number of grid points in the axis shape
      frenet (bool): True plots the Frenet-Serret frame, False it doesn't
      nphi_frenet (int): Number of Frenet-Serret vectors to show
      frenet_factor (float): Size of Frenet-Serret vectors
      savefig (string): filename to save resulting figure in png format.
        Note that ``.png`` will be appended.
        If ``None``, no figure file will be saved.
      show (bool): Whether or not to call the matplotlib/mayavi ``show()`` command.
    '''
    # Create array of toroidal angles along the axis
    # where the axis points will be created
    phi_array = np.linspace(0, 2 * np.pi, nphi)
    # Calculate the x, y and z components of the axis
    R0 = self.R0_func(phi_array)
    Z0 = self.Z0_func(phi_array)
    x_plot = R0 * np.cos(phi_array)
    y_plot = R0 * np.sin(phi_array)
    z_plot = Z0
    if frenet:
        # Show Frenet-Serret frame
        # Initiate mayavi instance
        from mayavi import mlab
        fig_axis = mlab.figure(bgcolor=(1,1,1), fgcolor=(0.,0.,0.), size=(650,500))
        # Plot the magnetic axis
        s = mlab.plot3d(x_plot, y_plot, z_plot, color=(0,0,0), line_width=0.001, tube_radius=0.01)
        # Show the x,y,z axis
        ax = mlab.axes(s,xlabel=r'X [m]',ylabel=r'Y [m]',zlabel=r'Z [m]',line_width=1.0,nb_labels=4)
        ax.axes.font_factor = 1.3
        ax.axes.label_format = '    %4.2f'
        ax.label_text_property.bold = False
        ax.label_text_property.italic = False
        # Create array of toroidal angles where the Frenet-Serret is shown
        phi_array = np.linspace(0, 2 * np.pi, nphi_frenet)
        # Calculate origin and vector arrays for the Frenet-Serret frame
        R0 = self.R0_func(phi_array)
        Z0 = self.Z0_func(phi_array)
        x_plot = R0*np.cos(phi_array)
        y_plot = R0*np.sin(phi_array)
        z_plot = Z0
        # Normal vector (red)
        normal_R   = self.normal_R_spline(phi_array)
        normal_phi = self.normal_phi_spline(phi_array)
        normal_Z   = self.normal_z_spline(phi_array)
        normal_X   = normal_R * np.cos(phi_array) - normal_phi * np.sin(phi_array)
        normal_Y   = normal_R * np.sin(phi_array) + normal_phi * np.cos(phi_array)
        mlab.quiver3d(x_plot, y_plot, z_plot,
                      normal_X, normal_Y, normal_Z,
                      scale_factor=frenet_factor,
                      color=(1, 0, 0), reset_zoom=False, name='Normal')
        # Biormal vector (blue)
        binormal_R   = self.binormal_R_spline(phi_array)
        binormal_phi = self.binormal_phi_spline(phi_array)
        binormal_Z   = self.binormal_z_spline(phi_array)
        binormal_X   = binormal_R * np.cos(phi_array) - binormal_phi * np.sin(phi_array)
        binormal_Y   = binormal_R * np.sin(phi_array) + binormal_phi * np.cos(phi_array)
        mlab.quiver3d(x_plot, y_plot, z_plot,
                      binormal_X, binormal_Y, binormal_Z,
                      scale_factor=frenet_factor,
                      color=(0, 0, 1), reset_zoom=False, name='Binormal')
        # Tangent vector (green)
        tangent_R   = self.tangent_R_spline(phi_array)
        tangent_phi = self.tangent_phi_spline(phi_array)
        tangent_Z   = self.tangent_z_spline(phi_array)
        tangent_X   = tangent_R * np.cos(phi_array) - tangent_phi * np.sin(phi_array)
        tangent_Y   = tangent_R * np.sin(phi_array) + tangent_phi * np.cos(phi_array)
        mlab.quiver3d(x_plot, y_plot, z_plot,
                      tangent_X, tangent_Y, tangent_Z,
                      scale_factor=frenet_factor,
                      color=(0, 1, 0),
                      reset_zoom=False, name='Tangent')
        if self.omn: # Show where curvature is zero if in quasi-isodynamism
            # zero_index = np.where(np.diff(self.sign_curvature_change)==-2)[0][0]
            # phi_at_zero = self.phi[zero_index]
            for i in range(2 * self.nfp):
                phi_at_zero = np.pi * i / self.nfp
                R0_at_zero = self.R0_func(phi_at_zero)
                Z0_at_zero = self.Z0_func(phi_at_zero)
                x_at_zero = R0_at_zero * np.cos(phi_at_zero)
                y_at_zero = R0_at_zero * np.sin(phi_at_zero)
                z_at_zero = Z0_at_zero
                mlab.points3d(x_at_zero, y_at_zero, z_at_zero, color=(0.9,0.9,0.9),scale_factor=0.09,figure=fig_axis)
        # Plot legends
        # Mayavi does not have its own legend function
        # Save figure
        if savefig != None:
            mlab.savefig(savefig + '.png')
        if show:
            # Show figure
            mlab.show()
    else:
        # Do not show Frenet-Serret frame
        # Initiate matplotlib instance
        fig_axis = plt.figure(figsize=(6, 5))
        ax = plt.axes(projection='3d')
        # Plot the magnetic axis
        plt.plot(x_plot, y_plot, z_plot)
        set_axes_equal(ax)
        ax.grid(False)
        ax.set_xlabel('X [m]', fontsize=10)
        ax.set_ylabel('Y [m]', fontsize=10)
        ax.set_zlabel('Z [m]', fontsize=10)
        plt.tight_layout()
        fig_axis.subplots_adjust(left=-0.05, top=1.05)
        # Save figure
        if savefig != None:
            fig_axis.savefig(savefig + '.pdf')
        if show:
            # Show figure
            plt.show()
